#include "cpsp/gecode/GC_HPThreading.hh"
#include "cpsp/gecode/GC_ThreadingSpace.hh"
#include "cpsp/HCoreDatabase.hh"
#include "biu/Timer.hh"

#include <iostream>

namespace cpsp
{
  namespace gecode
  {
	GC_HPThreading::GC_HPThreading(HPThreadingOptions * const opts_) :
		opts(opts_)
	{
	}
	
	GC_HPThreading::~GC_HPThreading()
	{
	}
	
	void
	GC_HPThreading::generateStructures(void) {
		HCore core;
		HCoreDatabase *db = opts->getHCoreDB();
		unsigned int sols = 0, oldSols = 0, oldContacts = INT_MAX,
			oldContactSols = 0;
		
		std::cout <<"==> HP-sequence = " <<*(opts->getSequence()) <<std::endl;
		
		if (opts->verboseOut())
			std::cout << "\ngenerating structures via Gecode\n" << std::endl;
		
		Gecode::Search::Statistics globStat;
			// timer setzen
		biu::Timer timer; timer.start();
		
		GC_ThreadingSymmBreaker::GlobalShiftVec* symmetryShiftVec = NULL;
		while (sols < opts->getMaximalSolutions() && db->getNextCore(core)) {
			try {
				if (opts->verboseOut()) {
					std::cout <<" => next core ("<<core.getContacts()
						<<" contacts, "
						<<core.getNumOfPSinglets(opts->getLattice()) 
						<<" P-singlets) "
						<<std::endl;
				}
					// naechster, wenn nicht genug singlet-positionen vorhanden
				if ( core.getNumOfPSinglets(opts->getLattice()) 
							< (int) opts->getNumOfPSinglets()
					|| core.getNumOfHSinglets(opts->getLattice()) 
							< (int) opts->getNumOfHSinglets()) 
				{
					if (opts->verboseOut()) {
						std::cout	<<"     --> skipped (singlet positions)"
									<<std::endl;
					}
					continue;
				}
					// naechster, wenn cubisches gitter und falsches verhaeltnis
					// zwischen gerade und ungeraden kernpositionen
				if ( opts->getLattice()->getDescriptor()->getName() == "cub"
					&& core.getMinEvenOdd() != opts->getMinEvenOddHs() ) 
				{
					if (opts->verboseOut()) {
						std::cout	<<"     --> skipped (cubic even/odd ratio) "
									<<std::endl;
					}
					continue;
				}
				
					// set minimal frame size 
				opts->setLatticeFrameSize(	opts->getMaxPCoreDistance() 
											+ core.getMaxDimension());

					// generate global shift vectors for symmetry breaking
				symmetryShiftVec = GC_ThreadingSymmBreaker::generateGlobalShiftVec(
										&core, opts->getLattice(), symmetryShiftVec);
										
					// s = pointer auf space object
				GC_ThreadingSpace* s = new GC_ThreadingSpace(opts, &core, symmetryShiftVec);
					// e = suchengine
				typedef Gecode::DFS<GC_ThreadingSpace> Engine;
				
				Engine dfs_e( s );
				
				delete s;

					// suche nach loesungen
				oldSols = sols;
				GC_ThreadingSpace* ex = NULL;
					// jeweils ein neuer space als loesung zurueckgegeben
				do {
						// ausgabe der energieniveaus
					if (	sols == (oldSols+1) 
							&& core.getContacts() < oldContacts) 
					{
						if (oldSols > 0) {
							std::cout	<<"\n ==> number generated structures = " 
										<<(oldSols - oldContactSols)
										<<std::endl;
								// number of solutions since last energy level change
							oldContactSols = oldSols;
						}
						
						oldContacts = core.getContacts();
						std::cout	<<"\n==> energy = -" 
									<<(oldContacts - opts->getContactsInSequence())
									<<std::endl;
					}

					ex = dfs_e.next();	// naechsten holen
					if (ex == NULL)	// keine weitere loesung
						break;
						
						// ausgabe der struktur, wenn noetig
					if (!opts->countingMode()) {
						std::cout	<<opts->getLattice()->getString(
										opts->getLattice()->indicesToAbsMoves(
											ex->getSolution()))
									<<std::endl;
					}
					delete ex;		// aufraeumen
				} while (++sols < opts->getMaximalSolutions());
				
				if (opts->verboseOut()) {
						// statistik daten holen
					Gecode::Search::Statistics actStat = dfs_e.statistics();
						// globale aktualisieren
					globStat.clone		+= actStat.clone;
					globStat.fail		+= actStat.fail;
					globStat.commit		+= actStat.commit;
					globStat.propagate	+= actStat.propagate;
					globStat.memory		+= actStat.memory;
					
					std::cout	<<"   + " <<(sols - oldSols) 
								<<" structures" <<std::endl;
				}
			} catch (Gecode::Exception e) {
				std::cout << "Exception in Gecode: " 
				 << e.what() << "." << std::endl
				<< "Stopping..." << std::endl;
			}
			
				// wenn nur beste gesucht werden, dann u.u. energieschranke 
				// anpassen
			if (	opts->getCoreSelection() == HPThreadingOptions::ALL_BEST 
					&& sols > 0) {
				db->setActMinHHcontacts(core.getContacts());
			}
		}
		if (symmetryShiftVec != NULL) delete symmetryShiftVec;		
		std::cout	<<"\n ==> number generated structures = " 
					<<(sols - oldContactSols)
					<<"\n\n- total number of generated structures = " 
					<<sols 
					<<std::endl;
					
		if (opts->verboseOut()) {
				// statistik ausgabe
			std::cout << "- statistics" << std::endl
			<< "\tsolutions:    " << sols << std::endl
			<< "\tpropagations: " << globStat.propagate << std::endl
			<< "\tfailures:     " << globStat.fail << std::endl
			<< "\tclones:       " << globStat.clone << std::endl
			<< "\tcommits:      " << globStat.commit << std::endl
			<< "\tpeak memory:  "
				<< static_cast<int>((globStat.memory+1023) / 1024) << " KB" 
				<< std::endl
			<< "\truntime:      " << timer.stop() <<" ms"
			<< std::endl;
		}
	}
	
	
  } // namespace gecode
} // namespace cpsp
